# Copyright contributors to the Terratorch project

from terratorch.models.heads.classification_head import ClassificationHead
from terratorch.models.heads.regression_head import RegressionHead
from terratorch.models.heads.segmentation_head import SegmentationHead

__all__ = ["ClassificationHead", "RegressionHead", "SegmentationHead"]
